#!/usr/bin/env python

from isicarchive import IsicApi

if __name__ == '__main__':
    
    # Create and print info about IsicApi object
    api = IsicApi('weberj3@mskcc.org', cache_folder='C:\\Users\\weberj3\\Documents\\git\\ISIC\\cache')
    api.cache_images()
    
    # Load and print info about a study
    study = api.study('EASY Study - 75 Image, full featureset')
    annotation = api.annotation(study.annotations[0])
    annotation._repr_pretty_()
    
    study.load_images()
    study.load_annotations()
    study.loaded_features

    # Retrieve an image for this study
    image = api.image(study.images[0])
    
    # Load the image data (from cache or online)
    image.load_imagedata()
    
    # And load the superpixels, and parse into coordinates map
    image.load_superpixels()
    image.map_superpixels()



"""
isicarchive.image

This module provides the Image object for the IsicApi to utilize.

Image objects are either returned from calls to

   >>> api = IsicApi()
   >>> image = api.image(image_id)

or can be generated

   >>> image = Image(...)
"""

__version__ = '0.4.0'


import datetime
import glob
import json
import os
import warnings

import imageio
import numpy
import requests

from . import func

_json_full_fields = [
    'created',
    'creator',
    'dataset',
    'id',
    'meta',
    'name',
    'notes',
    'updated',
]
_json_partial_fields = [
    'id',
    'name',
    'updated',
]
_mangling = {
    'id': '_id',
}
_repr_pretty_list = {
    'id': 'id',
    'name': 'name',
    'dataset_id': 'dataset._id',
    'dataset_name': 'dataset.name',
    'meta_acquisition': 'meta.acquisition.{}',
    'meta_clinical': 'meta.clinical.{keys}',
    'meta_clinical_benign_malignant': 'meta.clinical.benign_malignant',
    'meta_clinical_diagnosis': 'meta.clinical.diagnosis',
    'meta_clinical_diagnosis_confirm_type': 'meta.clinical.diagnosis_confirm_type',
    'meta_clinical_melanocytic': 'meta.clinical.melanocytic',
    'notes_reviewed': 'notes.reviewed.{}',
    'superpixels_max': 'superpixels.max',
    'superpixels_shape': 'superpixels.shp',
}

class Image(object):
    """
    Image object. If the details are not filled in, only the `id`, `name`,
    and `updated` fields will be set.

    To generate an image object for an existing dataset, please use the
    IsicApi.image(...) method!

    To generate a new image object (for later storage), use

       >>> image = Image(name=study_name, data=...)

    Attributes
    ----------
    created : Date
        Study creation date (w.r.t. in the database!)
    creator : dict
        Contains _id and (short) name field identifying the creator
    data : dict
        Once data is loaded, the imread decoded image will be in this field
    dataset : dict
        Dataset fields (non-detailed)
    id : str
        mongodb objectId of the study
    meta : dict
        Metadata associated with the image, containing fields
        'acquisition': dict (image type, size)
        'clinical': dict (patient/lesion classification)
        'private': dict (original filename, etc.)
        'unstructured': dict (additional metadata fields)
        'unstructuredExif': dict (additional image metadata fields)
    name : str
        ISIC archive filename (generally ISIC_#######, without extension!)
    notes : dict
        Notes (settings) within the archive, contains:
        'reviewed': dict (when was the image reviewed)
    updated : Date
        Image update date (w.r.t. in the database!)
    
    Methods
    -------
    """


    def __init__(self,
        from_json:dict = None,
        name:str = None,
        api:object = None,
        load_imagedata:bool = False,
        load_superpixels:bool = False,
        ):
        """Image init."""

        self._api = api
        self._detail = False
        self._in_archive = False
        # still needs timezone information!!
        self.created = datetime.datetime.now().strftime(
            '%Y-%m-%dT%H:%M:%S.%f+00:00')
        self.creator = {'_id': '000000000000000000000000'}
        self.data = None
        self.dataset = dict()
        self.id = ''
        self.meta = dict()
        self.name = name if name else 'ISIC_xxxxxxxx'
        self.notes = dict()
        self.superpixels = {
            'idx': None,
            'map': None,
            'max': 0,
            'shp': (0, 0),
        }
        self.updated = self.created

        # from JSON
        if isinstance(from_json, dict):
            try:
                self._from_json(from_json)
            except:
                raise
        if self._in_archive and load_imagedata:
            try:
                self.load_imagedata()
            except Exception as e:
                warnings.warn(str(e))
        if self._in_archive and load_superpixels:
            try:
                self.load_superpixels(map_superpixels=True)
            except Exception as e:
                warnings.warn(str(e))

    # parse JSON
    def _from_json(self, from_json:dict):
        self.id = from_json['_id']
        self.name = from_json['name']
        self.updated = from_json['updated']
        if 'creator' in from_json:
            self.created = from_json['created']
            self.creator = from_json['creator']
            if 'dataset' in from_json:
                self.dataset = from_json['dataset']
            if 'meta' in from_json:
                self.meta = from_json['meta']
            if 'notes' in from_json:
                self.notes = from_json['notes']
            self._detail = True
        self._in_archive = True

    # JSON
    def __repr__(self):
        return 'isicarchive.image.Image(from_json=%s)' % (self.as_json())
    
    # formatted print
    def __str__(self):
        return 'ISIC Image "{0:s}" (id={1:s}, image data loaded: {2:s})'.format(
            self.name, self.id, str(not self.data is None))
    
    # pretty print
    def _repr_pretty_(self, p:object, cycle:bool = False):
        func.object_pretty(self, p, cycle, _repr_pretty_list)
    
    # JSON representation (without constructor):
    def as_json(self):
        json_list = []
        fields = _json_full_fields if self._detail else _json_partial_fields
        for field in fields:
            if field in _mangling:
                json_field = _mangling[field]
            else:
                json_field = field
            json_list.append('"%s": %s' % (json_field,
                json.dumps(getattr(self, field))))
        return '{' + ', '.join(json_list) + '}'

    # load image data
    def load_imagedata(self):
        if not self._api:
            raise ValueError('Invalid image object to load image data for.')
        if self._api._cache_folder:
            image_filename = func.cache_filename(self.id, 'image', '.*', '*', api=self._api)
            image_list = glob.glob(image_filename)
            if image_list:
                try:
                    self.data = None
                    with open(image_list[0], 'rb') as image_file:
                        image_raw = image_file.read()
                    self.data = imageio.imread(image_raw)
                    return
                except Exception as e:
                    warnings.warn('Error loading image: ' + str(e))
                    os.remove(image_list[0])
        if self._in_archive and self._api._base_url:
            try:
                if self._api._auth_token:
                    headers = {'Girder-Token': self._api._auth_token}
                else:
                    headers = None
                req = requests.get(func.make_url(self._api._base_url,
                    'image/' + self.id + '/download'),
                    headers=headers, allow_redirects=True)
                if req.ok:
                    image_raw = req.content
                    self.data = imageio.imread(image_raw)
                    if self._api._cache_folder:
                        extra = self.name if (self.name and len(self.name) > 5) else None
                        image_filename = func.cache_filename(self.id, 'image',
                            func.guess_file_extension(req.headers), extra, api=self._api)
                        with open(image_filename, 'wb') as image_file:
                            image_file.write(image_raw)
            except Exception as e:
                warnings.warn('Error loading image data: ' + str(e))

    # load image superpixels
    def load_superpixels(self, map_superpixels:bool = False):
        if not self._api:
            raise ValueError('Invalid image object to load superpixels for.')
        spimg_filename = func.cache_filename(self.id, 'spimg', '.png', api=self._api)
        spidx_filename = func.cache_filename(self.id, 'spidx', '.npz', api=self._api)
        if self._api._cache_folder:
            if os.path.exists(spidx_filename):
                try:
                    if self.superpixels['idx'] is None:
                        spidx_data = numpy.load(spidx_filename)
                        if 'idx' in spidx_data:
                            self.superpixels['idx'] = spidx_data['idx']
                            self.superpixels['max'] = numpy.amax(
                                self.superpixels['idx']).item()
                            self.superpixels['shp'] = self.superpixels['idx'].shape
                except Exception as e:
                    os.remove(spidx_filename)
                    warnings.warn('Error loading spidx cache file: ' + str(e))
            if self.superpixels['idx'] is None:
                if os.path.exists(spimg_filename):
                    try:
                        with open(spimg_filename, 'rb') as image_file:
                            image_raw = image_file.read()
                        image_png = imageio.imread(image_raw)
                        self.superpixels['idx'] = func.superpixel_index(image_png)
                        self.superpixels['max'] = numpy.amax(
                            self.superpixels['idx']).item()
                        self.superpixels['shp'] = self.superpixels['idx'].shape
                        numpy.savez_compressed(spidx_filename, idx=self.superpixels['idx'])
                    except Exception as e:
                        os.remove(spimg_filename)
                        warnings.warn('Error loading image: ' + str(e))
        if self._in_archive and self._api._base_url and (not os.path.exists(spidx_filename)):
            try:
                if self._api._auth_token:
                    headers = {'Girder-Token': self._api._auth_token}
                else:
                    headers = None
                req = requests.get(func.make_url(self._api._base_url,
                    'image/' + self.id + '/superpixels'),
                    headers=headers, allow_redirects=True)
                if req.ok:
                    image_raw = req.content
                    image_png = imageio.imread(image_raw)
                    if self._api._cache_folder:
                        with open(spimg_filename, 'wb') as image_file:
                            image_file.write(image_raw)
                    self.superpixels['idx'] = func.superpixel_index(image_png)
                    self.superpixels['max'] = numpy.amax(
                        self.superpixels['idx']).item()
                    self.superpixels['shp'] = self.superpixels['idx'].shape
                    if self._api._cache_folder:
                        numpy.savez_compressed(spidx_filename,
                        idx=self.superpixels['idx'])
            except Exception as e:
                warnings.warn('Error loading superpixels: ' + str(e))
                return
        if not map_superpixels:
            return
        try:
            if self.superpixels['map'] is None:
                self.map_superpixels()
        except Exception as e:
            warnings.warn('Error processing superpixel data: ' + str(e))

    # map superpixels
    def map_superpixels(self):
        if not self.superpixels['map'] is None:
            return
        try:
            if self.superpixels['idx'] is None:
                self.load_superpixels()
            if self.superpixels['idx'] is None:
                raise ValueError('Some problem occurred during load_superpixels().')
        except Exception as e:
            warnings.warn('Error loading superpixels: ' + str(e))
            return
        pixel_img = self.superpixels['idx']
        try:
            self.superpixels['map'] = func.superpixel_decode_img(pixel_img)
        except Exception as e:
            warnings.warn('Error mapping superpixels: ' + str(e))

    # POST metadata for an image to the /image/{id}/metadata endpoint
    def post_metadata(self,
        metadata:dict,
        ) -> bool:
        """
        POSTs metadata for an image and returns True or False
        """
        if (not self._api) or (not self._api._auth_token):
            raise ValueError('Invalid image object to post metadata with.')
        if not func.could_be_mongo_object_id(self.id):
            raise ValueError('Invalid image object_id format.')
        url = func.make_url(self._api._base_url, 'image/' + self.id + '/metadata')
        req = requests.post(url,
            params={'metadata': metadata, 'save': 'true'},
            headers={'Girder-Token': self._api._auth_token})
        if not req.ok:
            warnings.warn("Image metadata posting failed: " + req.text)
            return ''
        return req.json()



"""
isicarchive variables

constants
---------
ISIC_API_URI : str
    current API URI
ISIC_BASE_URL : str
    hostname of ISIC Archive, including https:// protocol id
"""

__version__ = '0.4.0'

# Api URL/URI
ISIC_API_URI = '/api/v1'
ISIC_BASE_URL = 'https://isic-archive.com'

# Api image cache settings
ISIC_IMAGE_CACHE_UPDATE_LASTS = 900.0 # minimum time between updates in seconds
ISIC_IMAGES_PER_CACHING = 500 # number of image detail items per get(...) call



# isicarchive package
"""
ISIC Archive API

Provides a python interface to https://isic-archive.com/api/v1

To instantiate the API object and retrieve some study information use

   >>> from isicarchive import IsicApi
   >>> api = IsicApi(username)
   >>> studyInfo = api.study(study_name)

:copyright: (c) 2019, MSKCC.
:license: MIT, see LICENSE for details.
"""

import requests
import warnings

def check_dep_versions():
    r_version = requests.__version__.split('.')
    r_major = int(r_version[0])
    r_minor = int(r_version[1])
    if r_major < 2 or (r_major < 3 and r_minor < 22):
        warnings.warn("requests doesn't meet the minimum version requirement.")

check_dep_versions()

from .vars import __version__, ISIC_API_URI, ISIC_BASE_URL
from .api import IsicApi
from .annotation import Annotation
from .dataset import Dataset
from .image import Image
from .study import Study
from . import func

name = 'isicarchive'



"""
isicarchive.dataset

This module provides the Dataset object for the IsicApi to utilize.

Dataset objects are either returned from calls to

   >>> api = IsicApi()
   >>> dataset = api.dataset(dataset_id)

or can be generated

   >>> dataset = Dataset(...)
"""

__version__ = '0.4.0'


import datetime
import json
import warnings

import requests

from . import func
from .image import Image

_json_full_fields = [
    'access_level',
    'attribution',
    'count',
    'created',
    'creator',
    'description',
    'id',
    'license',
    'metadata_files',
    'name',
    'owner',
    'updated',
]
_json_partial_fields = [
    'access_level',
    'description',
    'id',
    'license',
    'name',
    'updated',
]
_mangling = {
    'access_level': '_accessLevel',
    'id': '_id',
    'metadata_files': 'metadataFiles',
}
_repr_pretty_list = [
    'id',
    'name',
    'description',
    'access_level',
    'license',
    'owner',
    'updated',
    'images',
    'images_for_review',
]

class Dataset(object):
    """
    Dataset object. If the details are not filled in, only the `id`,
    `description`, `name`, and `updated` fields will be set.

    To generate a study object for an existing study, please use the
    IsicApi.study(...) method!

    To generate a new study object (for later storage), use

       >>> study = Study(name=study_name, description=study_description)

    Attributes
    ----------
    access_level : Number
        Indicator for the kind of access level the dataset requires
    access_list : dict
        Access dict with access: {groups and users} and public: fields
    attribution : str
        Attribution for the dataset (e.g. author, organization)
    count : Number
        Number of authoritative/reviewed images in the dataset
    created : Date
        Study creation date (w.r.t. in the database!)
    creator : dict
        Contains _id and (short) name field identifying the creator
    description : str
        Longer description of study (optional)
    id : str
        mongodb objectId of the study
    images_for_review : list
        List of images to be reviewed
    license : str
        License short name (e.g. 'CC-0' or 'CC-BY-NC')
    metadata : any
        Metadata object as returned from ISIC Archive
    metadata_files : list
        List of metadata file references with fileId, time, and userId
    name : str
        Plain text name of the study (can be used for lookup also)
    owner : str
        Free text of the owner name (person or organization, not an id!)
    updated : Date
        Study update date (w.r.t. in the database!)
    
    Methods
    -------
    """


    def __init__(self,
        from_json:dict = None,
        name:str = None,
        description:str = None,
        api:object = None,
        ):
        """Dataset init."""

        self._api = api
        self._detail = False
        self._in_archive = False
        self._obj_images = dict()
        # still needs timezone information!!
        self.access_level = 2
        self.access_list = []
        self.attribution = 'Anonymous'
        self.count = 0
        self.created = datetime.datetime.now().strftime(
            '%Y-%m-%dT%H:%M:%S.%f+00:00')
        self.creator = {'_id': '000000000000000000000000'}
        self.description = description if description else ''
        self.id = ''
        self.images = []
        self.images_for_review = []
        self.license = 'CC-0'
        self.metadata = []
        self.metadata_files = []
        self.metadata_filecont = dict()
        self.name = name if name else ''
        self.owner = 'Anonymous'
        self.updated = self.created

        # preference: JSON, id (in name), then name (lookup)
        if isinstance(from_json, dict):
            try:
                self._from_json(from_json)
            except:
                raise
        elif func.could_be_mongo_object_id(self.name) and self._api and self._api._base_url:
            try:
                self._from_json(func.get(self._api._base_url,
                'dataset/' + self.name, self._api._auth_token).json())
            except:
                raise
        elif self.name and self._api and self._api._base_url:
            try:
                dataset_lookup = func.get(self._api._base_url,
                    'dataset', self._api._auth_token,
                    params={'limit': 0, 'detail': 'false'}).json()
                for dataset in dataset_lookup:
                    if dataset['name'] == dataset.name:
                        self._from_json(func.get(self._api._base_url,
                            'dataset/' + dataset['_id'],
                            self._api._auth_token).json())
                        break
                if not self.id:
                    warnings.warn('Dataset {0:s} not found.'.format(self.name))
            except:
                raise

    # parse JSON
    def _from_json(self, from_json:dict):
        self.access_level = from_json['_accessLevel']
        self.description = from_json['description']
        self.id = from_json['_id']
        self.license = from_json['license']
        self.name = from_json['name']
        self.updated = from_json['updated']
        if 'creator' in from_json:
            self.created = from_json['created']
            self.creator = from_json['creator']
            if 'attribution' in from_json:
                self.attribution = from_json['attribution']
            if 'count' in from_json:
                self.count = from_json['count']
            if 'metadataFiles' in from_json:
                self.metadata_files = from_json['metadataFiles']
            if 'owner' in from_json:
                self.owner = from_json['owner']
            self._detail = True
        self._in_archive = True
        if self._api and self._api._base_url and self._api._auth_token:
            try:
                self.access_list = func.get(self._api._base_url,
                    'dataset/' + self.id + '/access',
                    self._api._auth_token).json()
            except:
                warnings.warn('Error retrieving dataset access list.')
            try:
                self.metadata = func.get(self._api._base_url,
                    'dataset/' + self.id + '/metadata',
                    self._api._auth_token).json()
            except:
                warnings.warn('Error retrieving dataset metadata.')
            try:
                self.images_for_review = func.get(self._api._base_url,
                    'dataset/' + self.id + '/review',
                    self._api._auth_token, params={'limit': 0}).json()
                if not isinstance(self.images_for_review, list):
                    self.images_for_review = []
            except:
                warnings.warn('Error retrieving images for review.')

    # JSON
    def __repr__(self):
        return 'isicarchive.dataset.Dataset(from_json=%s)' % (self.as_json())
    
    # formatted print
    def __str__(self):
        return 'ISIC Dataset "{0:s}" (id={1:s}, {2:d} reviewed images)'.format(
            self.name, self.id, self.count)
    
    # pretty print
    def _repr_pretty_(self, p:object, cycle:bool = False):
        func.object_pretty(self, p, cycle, _repr_pretty_list)

    # JSON representation (without constructor):
    def as_json(self):
        json_list = []
        fields = _json_full_fields if self._detail else _json_partial_fields
        for field in fields:
            if field in _mangling:
                json_field = _mangling[field]
            else:
                json_field = field
            json_list.append('"%s": %s' % (json_field,
                json.dumps(getattr(self, field))))
        return '{' + ', '.join(json_list) + '}'

    # load images
    def load_images(self, load_imagedata:bool = False):
        if not self._api or not self._api._cache_folder:
            raise ValueError('Requires IsicApi object with cache folder set.')
        self._api.cache_images()
        dataset_id = self.id
        self.images = [item for item in self._api._image_cache.values() if (
            item['dataset']['_id'] == dataset_id)]
        if len(self.images) == 0:
            return
        for img_idx in range(len(self.images)):
            image_detail = self.images[img_idx]
            image_id = image_detail['_id']
            if image_id in self._api._image_objs:
                self._obj_images[image_id] = self._api._image_objs[image_id]
                continue
            image_obj = Image(from_json=image_detail,
                api=self._api, load_imagedata=load_imagedata)
            self._obj_images[image_id] = image_obj
            self._api._image_objs[image_id] = image_obj

    # POST an image to the /dataset/{id}/image endpoint
    def post_image(self,
        local_filename:str,
        signature:str,
        ) -> str:
        """
        POSTs an image (local file) to a dataset and returns the object_id
        """
        if (not self._api) or (not func.could_be_mongo_object_id(self.id)):
            raise ValueError('Invalid dataset object for image upload.')
        if ((local_filename is None) or (local_filename == '')):
            raise ValueError('Requires a non-empty filename.')
        if ((signature is None) or (signature == '')):
            raise ValueError('Requires a signature.')
        try:
            with open(local_filename, 'rb') as file_id:
                file_content = file_id.read()
        except:
            raise
        try:
            imageio.imread(file_content)
        except:
            raise ValueError('Error imread-ing image: ' + local_filename)
        url = func.make_url(self._api._base_url,
            'dataset/' + self.id + '/image')
        if self._api._auth_token:
            headers = {'Girder-Token': self._api._auth_token}
        else:
            headers = None
        req = requests.post(url,
            data=file_content,
            params={},
            headers=headers)
        if not req.ok:
            warnings.warn("Image upload failed: " + req.text)
            return ''
        msg = req.json()
        if '_id' in msg:
            object_id = msg['_id']
        else:
            warnings.warn('Cannot return image object: ' + req.text)
            return None
        return self._api.image(object_id)



"""
isicarchive.annotation

This module provides the Annotation object for the IsicApi to utilize.

Annotation objects are either returned from calls to

   >>> api = IsicApi()
   >>> study = api.study(study_name)
   >>> annotation = study.load_annotation(annotation_id)

or can be generated

   >>> annotation = Annotation(...)
"""

__version__ = '0.4.0'


import datetime
import json
import warnings

import numpy
import requests

from . import func

_json_full_fields = [
    'id',
    'image',
    'image_id',
    'log',
    'markups',
    'responses',
    'start_time',
    'state',
    'status',
    'stop_time',
    'study_id',
    'user',
    'user_id',
]
_mangling = {
    'id': '_id',
    'image_id': 'imageId',
    'start_time': 'startTime',
    'stop_time': 'stopTime',
    'study_id': 'studyId',
    'user_id': 'userId',
}
_repr_pretty_list = {
    'id': 'id',
    'study_id': 'study_id',
    'study_name': '_study.name',
    'image_id': 'image_id',
    'image_name': 'image.name',
    'user_id': 'user_id',
    'user_name': 'user.name',
    'user_lastname': 'user.lastName',
    'features': 'features.{keys}',
}

class Annotation(object):
    """
    Annotation object. If the details are not filled in, only the `id`,
    `image_id`, `state`, `study_id`, and `user_id` fields will be set.

    To generate an annotation object for an existing study, please use
    the Study.load_annotation(...) method!

    To generate a new annotation object (for later storage), use

       >>> anno = Annotation(image=image_id, study=study_id, user=user_id)

    Attributes
    ----------
    features : dict
        List of features (fields: idx and msk)
    id : str
        mongodb objectId of the annotation
    image : dict
        Minimal dict of the image (fields: _id, name, updated)
    image_id : str
        mongodb objectId of the image the annotation pertains to
    log : list
        Log entries (can be missing!)
    markups : dict
        Features that are part of the annotation {'Feature name': bool, ...}
        (can be missing!)
    responses : dict
        List of responses to questions {'Question': 'Response', ...}
        (can be missing!)
    start_time : Date
        Time when annotation was started (can be missing!)
    state : str
        State of annotation (one of 'active', 'complete' [?])
    status : str
        Status of annotation (can be missing!)
    stop_time : Date
        Time when annotation was finished (can be missing!)
    study_id : str
        mongodb objectId of the study the annotation pertains to
    user : dict
        Minimal dict of the user (fields: _id, name)
    user_id : str
        mongodb objectId of the user who entered the annotation
    
    Methods
    -------
    """


    def __init__(self,
        from_json:dict = None,
        annotation_id:str = None,
        image:str = None,
        study:str = None,
        user:str = None,
        api:object = None,
        load_data:bool = False,
        ):
        """Annotation init."""

        self._api = api
        self._in_archive = False
        self._study = None
        self.features = dict()
        self.id = ''
        self.image = None
        self.image_id = image
        self.log = []
        self.markups = dict()
        self.responses = dict()
        self.start_time = ''
        self.state = 'active'
        self.status = '?'
        self.stop_time = ''
        self.study_id = study
        self.user = None
        self.user_id = user

        # preference: JSON, id (in name), then name (lookup)
        if isinstance(from_json, dict):
            try:
                self._from_json(from_json, load_data)
            except:
                raise
        elif func.could_be_mongo_object_id(annotation_id) and self._api and self._api._base_url:
            try:
                self._from_json(func.get(self._api._base_url,
                'annotation/' + annotation_id, self._api._auth_token).json(), load_data)
            except:
                raise

    # parse JSON
    def _from_json(self, from_json:dict, load_data:bool = False):
        self.id = from_json['_id']
        self.state = from_json['state']
        self.study_id = from_json['studyId']
        if 'image' in from_json:
            self.image = from_json['image']
            self.image_id = self.image['_id']
        elif 'imageId' in from_json:
            self.image_id = from_json['imageId']
        if 'log' in from_json:
            self.log = from_json['log']
        if 'markups' in from_json:
            self.markups = from_json['markups']
        if 'responses' in from_json:
            self.responses = from_json['responses']
        if 'startTime' in from_json:
            self.start_time = from_json['startTime']
        if 'status' in from_json:
            self.status = from_json['status']
        if 'stopTime' in from_json:
            self.stop_time = from_json['stopTime']
        if 'user' in from_json:
            self.user = from_json['user']
            self.user_id = self.user['_id']
        elif 'userId' in from_json:
            self.user_id = from_json['userId']
        self._in_archive = True
        if self._api and self.study_id in self._api._studies:
            self._study = self._api._studies[self.study_id]
        if (not load_data) and (self.state != 'complete'):
            return
        if self._api and self._api._base_url and self._api._auth_token:
            try:
                headers = {'Girder-Token': self._api._auth_token}
                if ('features' in from_json and 
                    isinstance(from_json['features'], dict)):
                    self.features = from_json['features']
                for (key, value) in self.markups.items():
                    if not value:
                        continue
                    if not (key in self.features and
                        isinstance(self.features[key], dict) and
                        ('idx' in self.features[key]) and
                        (len(self.features[key]['idx']) > 0)):
                        feat_uri = func.uri_encode(key)
                        feat_lst = func.get(self._api._base_url,
                            'annotation/' + self.id + '/' + feat_uri,
                            self._api._auth_token)
                        if not feat_lst.ok:
                            raise ValueError('Error loading feature ' + key)
                        feat_lst = feat_lst.json()
                        feat_idx = numpy.flatnonzero(feat_lst).tolist()
                        self.features[key] = dict()
                        self.features[key]['lst'] = [v for v in filter(
                            lambda v: v > 0, feat_lst)]
                        self.features[key]['idx'] = feat_idx
                        self.features[key]['num'] = len(feat_idx)
                    if not load_data:
                        continue
                    feat_req = requests.get(
                        self._api._base_url + '/annotation/' + self.id +
                            '/' + feat_uri + '/mask', headers=headers,
                        allow_redirects=True)
                    if not feat_req.ok:
                        raise ValueError('Error loading feature mask ' + key)
                    self.features[key]['msk'] = feat_req.content
            except Exception as e:
                warnings.warn('Error loading annotation: ' + str(e))

    # JSON
    def __repr__(self):
        return 'isicarchive.annotation.Annotation(from_json=%s)' % (self.as_json())
    
    # formatted print
    def __str__(self):
        return 'ISIC Annotation (id={0:s}, image_id={1:s}, study_id={2:s})'.format(
            self.id, self.image_id, self.study_id)
    
    # pretty print
    def _repr_pretty_(self, p:object, cycle:bool = False):
        func.object_pretty(self, p, cycle, _repr_pretty_list)

    # JSON representation (without constructor):
    def as_json(self):
        json_list = []
        for field in _json_full_fields:
            if field in _mangling:
                json_field = _mangling[field]
            else:
                json_field = field
            json_list.append('"%s": %s' % (json_field,
                json.dumps(getattr(self, field))))
        return '{' + ', '.join(json_list) + '}'



"""
isicarchive.study

This module provides the Study object for the IsicApi to utilize.

Study objects are either returned from calls to

   >>> api = IsicApi()
   >>> study = api.study(study_id)

or can be generated

   >>> study = Study(...)
"""

__version__ = '0.4.0'

import copy
import datetime
import json
import numbers
import os
import warnings

from .annotation import Annotation
from .image import Image
from . import func

_images_per_request = 50
_json_full_fields = [
    'created',
    'creator',
    'description',
    'features',
    'id',
    'images',
    'name',
    'participation_requests',
    'questions',
    'updated',
    'user_completion',
    'users',
]
_json_partial_fields = [
    'description',
    'id',
    'name',
    'updated',
]
_mangling = {
    'id': '_id',
    'participation_requests': 'participationRequests',
    'user_completion': 'userCompletion',
}
_repr_pretty_list = [
    'id',
    'name',
    'description',
    'annotations',
    'features',
    'images',
    'questions',
    'users',
]

class Study(object):
    """
    Study object. If the details are not filled in, only the `description`,
    `id`, `name`, and `updated` fields will be set.

    To generate a study object for an existing study, please use the
    IsicApi.study(...) method!

    To generate a new study object (for later storage), use

       >>> study = Study(name=study_name, description=study_description)

    Attributes
    ----------
    created : Date
        Study creation date (w.r.t. in the database!)
    creator : dict
        Contains _id and (short) name field identifying the creator
    description : str
        Longer description of study (optional)
    features : list
        List of features associated with the study, {'id': 'Feature name'}
    id : str
        mongodb objectId of the study
    images : list
        List of images associated with a study,
        {'_id': object_id, 'name': 'ISIC_#######', 'updated': Date}
    name : str
        Plain text name of the study (can be used for lookup also)
    participation_requests : list
        List of participation requests by users
    questions : list
        List of questions associated with the study,
        {'choices': List[...], 'id': 'Question name', 'type', TYPE}
    updated : Date
        Study update date (w.r.t. in the database!)
    user_completion : dict
        Maps user object_ids to the number of completed images
    users : list
        List of users having taken the study,
        {'_id': object_id, 'name': 'User ####'}
    
    Methods
    -------
    annotation
        Return an annotation object associated with an object_id
    load_annotations
        Turn the annotation (JSON dict) list into objects
    """


    def __init__(self,
        from_json:dict = None,
        name:str = None,
        description:str = None,
        api:object = None,
        image_details:bool = True,
        ):
        """Study init."""

        self._annotations = dict()
        self._api = api
        self._detail = False
        self._in_archive = False
        self._obj_annotations = dict()
        self._obj_images = dict()
        # still needs timezone information!!
        self.annotations = []
        self.created = datetime.datetime.now().strftime(
            '%Y-%m-%dT%H:%M:%S.%f+00:00')
        self.creator = {'_id': '000000000000000000000000'}
        self.description = description if description else ''
        self.features = []
        self.id = ''
        self.images = []
        self.image_features = dict()
        self.loaded_features = dict()
        self.loaded_features_in = dict()
        self.name = name if name else ''
        self.participation_requests = []
        self.questions = []
        self.updated = self.created
        self.user_completion = dict()
        self.users = []

        # preference: JSON, id (in name), then name (lookup)
        if isinstance(from_json, dict):
            try:
                self._from_json(from_json, image_details)
            except:
                raise
        elif func.could_be_mongo_object_id(self.name) and self._api:
            try:
                self._from_json(func.get(self._api._base_url,
                    'study/' + self.name, self._api._auth_token).json(),
                    image_details)
            except:
                raise
        elif self.name and self._api:
            try:
                study_lookup = func.get(self._api._base_url,
                    'study', self._api._auth_token,
                    params={'limit': 0, 'detail': 'false'}).json()
                for study in study_lookup:
                    if study['name'] == self.name:
                        self._from_json(func.get(self._api._base_url,
                            'study/' + study['_id'], self._api._auth_token).json(),
                            image_details)
                        break
                if not self.id:
                    warnings.warn('Study {0:s} not found.'.format(self.name))
            except:
                raise

    # read from JSON
    def _from_json(self, from_json:dict, image_details:bool = True):
        self.description = from_json['description']
        self.id = from_json['_id']
        self.name = from_json['name']
        self.updated = from_json['updated']
        if 'creator' in from_json:
            self.created = from_json['created']
            self.creator = from_json['creator']
            if 'features' in from_json:
                self.features = from_json['features']
            if 'images' in from_json:
                self.images = from_json['images']
            if 'participationRequests' in from_json:
                self.participation_requests = from_json['participationRequests']
            if 'questions' in from_json:
                self.questions = from_json['questions']
            if 'userCompletion' in from_json:
                self.user_completion = from_json['userCompletion']
            if 'users' in from_json:
                self.users = from_json['users']
            self._detail = True
        self._in_archive = True
        if self._api and self._api._auth_token:
            try:
                if len(self.images) > 0 and image_details:
                    self.load_images()
            except:
                warnings.warn('Error retrieving image information.')
            try:
                annotations = func.get(self._api._base_url,
                    'annotation', self._api._auth_token,
                    params={'studyId': self.id, 'detail': 'true'}).json()
                self.annotations = annotations
                for count in range(len(annotations)):
                    self._annotations[annotations[count]['_id']] = count
            except:
                warnings.warn('Error retrieving annotations.')

    # JSON
    def __repr__(self):
        return 'isicarchive.study.Study(from_json=%s)' % (self.as_json())
    
    # formatted print
    def __str__(self):
        return 'ISIC Study "{0:s}" (id={1:s}, {2:d} images)'.format(
            self.name, self.id, len(self.images))
    
    # pretty print
    def _repr_pretty_(self, p:object, cycle:bool = False):
        func.object_pretty(self, p, cycle, _repr_pretty_list)
    
    # JSON representation (without constructor):
    def as_json(self):
        json_list = []
        fields = _json_full_fields if self._detail else _json_partial_fields
        for field in fields:
            if field in _mangling:
                json_field = _mangling[field]
            else:
                json_field = field
            json_list.append('"%s": %s' % (json_field,
                json.dumps(getattr(self, field))))
        return '{' + ', '.join(json_list) + '}'

    # get annotation
    def annotation(self, object_id:str):
        if not self._api:
            raise ValueError('Requires IsicApi object to be set.')
        if isinstance(object_id, numbers.Number) and (
            object_id >= 0 and object_id < len(self.annotations)):
                object_id = self.annotations[object_id]['_id']
        if isinstance(object_id, str) and func.could_be_mongo_object_id(object_id):
            if object_id in self._obj_annotations:
                return self._obj_annotations[object_id]
            if not object_id in self._annotations:
                raise ValueError('Invalid or missing object_id in call')
            if self._api and object_id in self._api._annotation_objs:
                annotation_obj = self._api._annotation_objs[object_id]
                self._obj_annotations[object_id] = annotation_obj
                return annotation_obj
            object_id = self.annotations[self._annotations[object_id]]
        if not isinstance(object_id, dict) or (
            not '_id' in object_id or
            not '_modelType' in object_id or
            object_id['_modelType'] != 'annotation'):
            raise ValueError('Invalid or missing object_id in call')
        try:
            annotation_obj = Annotation(from_json = object_id, api=self._api)
            self._obj_annotations[object_id['_id']] = annotation_obj
            self._api._annotation_objs[object_id['_id']] = annotation_obj
        except:
            raise
        for key, value in annotation_obj.features.items():
            try:
                if isinstance(value, dict) and ('lst' in value) and value['lst']:
                    if not key in self.loaded_features:
                        self.loaded_features[key] = 0
                    self.loaded_features[key] += 1
                    if not key in self.loaded_features_in:
                        self.loaded_features_in[key] = list()
                    self.loaded_features_in[key].append(object_id['_id'])
            except:
                warnings.warn(
                    'Error adding feature {0:s} to list for annotation id={1:s}.'.format(
                    key, object_id['_id']))
        return annotation_obj

    # image names
    def image_names(self):
        return [image['name'] for image in self.images]

    # load annotations
    def load_annotations(self):
        if (not self._api) or len(self._obj_annotations) == len(self.annotations):
            return
        study_anno_filename = func.cache_filename(self.id,
            'stann', '.json.gz', api=self._api)
        study_anno_data = dict()
        if os.path.exists(study_anno_filename):
            try:
                study_anno_data = func.gzip_load_var(study_anno_filename)
            except Exception as e:
                os.remove(study_anno_filename)
                warnings.warn('Error reading study annotations file: ' + str(e))
        didwarn = False
        for annotation in self.annotations:
            annotation_id = annotation['_id']
            if not annotation_id in self._obj_annotations:
                if annotation_id in study_anno_data:
                    annotation = copy.copy(annotation)
                    annotation['features'] = study_anno_data[annotation_id]
                try:
                    self.annotation(annotation)
                    annotation_obj = self._obj_annotations[annotation_id]
                    annotation_features = annotation_obj.features
                    for key, val in annotation_obj.markups.items():
                        if not val:
                            continue
                        if not annotation_id in study_anno_data:
                            study_anno_data[annotation_id] = dict()
                        study_anno_data[annotation_id][key] = annotation_features[key]
                except Exception as e:
                    warnings.warn('Error retrieving annotation {0:s} details: {1:s}'.format(
                        annotation['_id'], str(e)))
                    didwarn = True
        if not didwarn:
            if os.path.exists(study_anno_filename):
                os.remove(study_anno_filename)
            func.gzip_save_var(study_anno_filename, study_anno_data)


    # load images
    def load_images(self, load_imagedata:bool = False):
        if (not self._api) or (len(self.images) == 0):
            return
        params = {
            'detail': 'true',
            'imageIds': '',
        }
        to_load = []
        rep_idx = dict()
        for count in range(len(self.images)):
            image_id = self.images[count]['_id']
            if image_id in self._api._image_cache:
                if image_id in self._api._image_objs:
                    self._obj_images[image_id] = self._api._image_objs[image_id]
                    continue
                image_detail = self._api._image_cache[image_id]
                image_obj = Image(from_json=image_detail,
                    api=self._api, load_imagedata=load_imagedata)
                self._obj_images[image_id] = image_obj
                self._api._image_objs[image_id] = image_obj
                continue
            if not '_modelType' in self.images[count]:
                to_load.append(image_id)
                rep_idx[image_id] = count
            if len(to_load) == _images_per_request:
                params['imageIds'] = '["' + '","'.join(to_load) + '"]'
                image_info = func.get(self._api._base_url,
                    'image', self._api._auth_token, params=params).json()
                if len(image_info) != len(to_load):
                    raise ValueError('Bad request output.')
                for repcount in range(len(image_info)):
                    image_detail = image_info[repcount]
                    image_id = image_detail['_id']
                    self.images[rep_idx[image_id]] = image_detail
                    if image_id in self._api._image_objs:
                        self._obj_images[image_id] = self._api._image_objs[image_id]
                        continue
                    image_obj = Image(from_json=image_detail,
                        api=self._api, load_imagedata=load_imagedata)
                    self._obj_images[image_id] = image_obj
                    self._api._image_objs[image_id] = image_obj
                to_load = []
                rep_idx = dict()
        if len(to_load) > 0:
            params['imageIds'] = '["' + '","'.join(to_load) + '"]'
            image_info = func.get(self._api._base_url,
                'image', self._api._auth_token, params=params).json()
            if len(image_info) != len(to_load):
                raise ValueError('Bad request output.')
            for repcount in range(len(image_info)):
                image_detail = image_info[repcount]
                image_id = image_detail['_id']
                self.images[rep_idx[image_id]] = image_detail
                if image_id in self._api._image_objs:
                    self._obj_images[image_id] = self._api._image_objs[image_id]
                    continue
                image_obj = Image(from_json=image_detail,
                    api=self._api, load_imagedata=load_imagedata)
                self._obj_images[image_id] = image_obj
                self._api._image_objs[image_id] = image_obj



#!/usr/bin/env python

import setuptools

from isicarchive.vars import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

requires = [
    'imageio>=2.5.0',
    'numba>=0.45.1',
    'numpy>=1.16.2',
    'requests>=2.22.0',
]

setuptools.setup(
    name="isicarchive",
    version=__version__,
    description="ISIC Archive API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jochen Weber",
    author_email="weberj3@mskcc.org",
    url="https://github.com/neuroelf/isicarchive",
    packages=setuptools.find_packages(),
    package_dir={'isicarchive': 'isicarchive'},
    python_requires=">=3.6",
    install_requires=requires,
    license='MIT',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)



"""
isicarchive.func

This module provides helper functions and doesn't have to be
imported from outside the main package functionality (isicapi).

Functions
---------
cache_filename
    Creates a filename of a specific type from an id
could_be_mongo_object_id
    Returns true if the input is a 24 lower-case hex character string
get
    Specific URL mangling rules prior to calling requests.get(...)
get_json
    Passes the input through get(...) and appends .json()
get_json_list
    Passes the input through get(...) and yields one array item
guess_file_extension
    Guesses a downloaded file's extension from the HTTP Headers
gzip_load_var
    Loads a .json.gz file into a variable
gzip_save_var
    Saves a variable into a .json.gz file
isic_auth_token
    Makes a login attempt and extracts the Girder-Token from the Headers
make_url
    Concatenates URL particles
superpixel_decode_img
    Decodes a superpixel (index) array into a map (dict)
superpixel_index
    Converts an RGB superpixel image to a 2D superpixel index array
uri_encode
    Encodes non-letter/number characters into %02x sequences
"""

__version__ = '0.4.0'

import copy
import gzip
import json
import os
import re
from typing import Any, Tuple
import warnings

import numba
import numpy
import requests

# cache filename
def cache_filename(oid:str = None, otype:str = None, oext:str = None, extra:str = None, api:object = None) -> str:
    """
    Creates a filename out of an object (type, id, ext).

    Parameters
    ----------
    oid : str
        object_id (mongodb objectId)
    otype : str
        Arbitrary object type str (first part of filename)
    oext : str
        Arbitrary extension, will be .bin if none
    extra : str
        String folder in before extension
    api : IsicApi
        Necessary to access cache folder, if api not valid, returns None
    
    Returns
    -------
    filename : str
        Filename of object in cache
    """
    if api is None or api._cache_folder is None:
        return None
    if not isinstance(oid, str) or (not could_be_mongo_object_id(oid)):
        raise ValueError('Invalid object_id')
    if otype is None or (otype == ''):
        otype = 'object'
    if oext is None or (oext == ''):
        oext = '.bin'
    elif oext[0] != '.':
        oext = '.' + oext
    if (not extra is None) and (extra != ''):
        extra = '_' + extra
    else:
        extra = ''
    return (api._cache_folder +
        os.sep + oid[-1] + os.sep + otype + '_' + oid + extra + oext)

# helper function that returns True for valid looking mongo ObjectId strings
_mongo_object_id_pattern = re.compile(r"^[0-9a-f]{24}$")
def could_be_mongo_object_id(test_id:str = "") -> bool:
    """
    Tests if passed-in string is 24 lower-case hexadecimal characters.

    Parameters
    ----------
    test_id : str
        String representing a possible mongodb objectId
    
    Returns
    -------
    test_val : bool
        True if test_id is 24 characters of lower-case hexadecimal characters
    """
    return (len(test_id) == 24
            and (not re.match(_mongo_object_id_pattern, test_id) is None))

# Generic endpoint API, allowing arbitrary commands
def get(
    base_url:str,
    endpoint:str,
    auth_token:str = None,
    params:dict = None,
    save_as:str = None,
    ) -> Any:
    """
    Performs a GET request to the given endpoint, with the provided
    parameters. If the `save_as` argument is given, will attempt to
    store the returned output into a local file instead of returning
    the content as a string.

    Parameters
    ----------
    base_url : str
        Base URL from which to build the full URL
    endpoint : str
        Endpoint (URI) to which the request is made, WITHOUT leading /
    auth_token : str
        Girder-Token, which is sent as a header (or None)
    params : dict
        Optional parameters that will be added to the query string
    save_as : str
        Optional string containing a local target filename
    
    Returns
    -------
    content : Any
        If the request is successful, the returned content
    """

    url = make_url(base_url, endpoint)
    headers = {'Girder-Token': auth_token} if auth_token else None
    if save_as is None:
        return requests.get(url, headers=headers, params=params)
    req = requests.get(url,
        headers=headers,
        params=params,
        allow_redirects=True)
    open(save_as, 'wb').write(req.content)

# Generic endpoint that already converts content to JSON
def get_json(
    base_url:str,
    endpoint:str,
    auth_token:str = None,
    params:dict = None,
    ) -> object:
    """
    Performs a GET request and parses the returned content using the
    requests.get(...).json() method.

    Passes through `get(...)`, see parameters there.
    """
    return get(base_url, endpoint, auth_token, params).json()

# Generic endpoint to generate iterator over JSON list
def get_json_list(
    base_url:str,
    endpoint:str,
    auth_token:str = None,
    params:dict = None,
    ) -> iter:
    """
    Generates an iterator to handle one (JSON) item at a time.

    For syntax, see `self.get(...)`

    Yields
    ------
    object
        one JSON object from an array
    """
    resp = get(base_url, endpoint, auth_token, params).json()
    for item in resp:
        yield item

def getxattr(obj:object, name:str = None, default:Any = None) -> Any:
    """
    Get attribute or key-based value from object

    Parameters
    ----------
    obj : object
        Either a dictionary or object with attributes
    name : str
        String describing what to retrieve
    
    Returns
    -------
    value : Any
        Value from obj.name where name can be name1.name2.name3
    """
    val = default
    if obj is None:
        return val
    if not '.' in name:
        try:
            if isinstance(obj, dict):
                val = obj.get(name)
            elif isinstance(obj, list) and name.isdigit():
                val = obj[int(name)]
            else:
                val = getattr(obj, name)
        except:
            pass
        return val
    name_lst = name.split('.')
    name_lst.reverse()
    try:
        while len(name_lst) > 1:
            obj = getxattr(obj, name_lst.pop())
            if obj is None:
                return val
        if isinstance(obj, list) and (name_lst[0] == '[]'):
            val = '[' + ', '.join([repr(x) for x in obj]) + ']'
        elif isinstance(obj, dict) and (name_lst[0] == '{keys}'):
            val = '{' + ', '.join([repr(x) for x in obj.keys()]) + '}'
        elif isinstance(obj, dict) and (name_lst[0] == '{}'):
            val = '{' + ', '.join(
                [repr(k) + ': ' + repr(v) for k,v in obj.items()]) + '}'
        else:
            val = getxattr(obj, name_lst[0])
    except:
        pass
    return val

# guess file extentions from returned request headers
_ext_type_guess = {
    'bmp': '.bmp',
    'gif': '.gif',
    'jpeg': '.jpg',
    'jpg': '.jpg',
    'png': '.png',
}
def guess_file_extension(headers:dict) -> str:
    """
    Guesses the file extension of a requests.get content from its headers

    Parameters
    ----------
    headers : dict
        Headers as available in requests.get(...).headers

    Returns
    -------
    file_ext : str
        File extension guess (including leading dot!), or '.bin' otherwise
    """
    ctype = None
    cdisp = None
    if 'Content-Type' in headers:
        ctype = headers['Content-Type']
    elif 'content-type' in headers:
        ctype = headers['content-type']
    if ctype:
        ctype = ctype.split('/')
        ctype = _ext_type_guess.get(ctype[-1], None)
    if ctype:
        return ctype
    if 'Content-Disposition' in headers:
        cdisp = headers['Content-Disposition']
    elif 'content-disposition' in headers:
        cdisp = headers['content-disposition']
    if cdisp:
        if 'filename=' in cdisp.lower():
            filename = cdisp.split('ilename=')
            filename = filename[-1]
            if filename[0] in r'\'"' and filename[0] == filename[-1]:
                filename = filename[1:-1]
            filename = filename.split('.')
            if filename[-1].lower() in _ext_type_guess:
                return _ext_type_guess[filename[-1].lower()]
    return '.bin'

# load JSON.gz GZIP file into variable
def gzip_load_var(gzip_file:str) -> Any:
    """
    Load variable from .json.gz file (arbitrary extension!)

    Parameters
    ----------
    gzip_file : str
        Filename containing the gzipped JSON variable
    
    Returns
    -------
    var : Any
        Variable as decoded from gzipped JSON content
    """
    try:
        with gzip.GzipFile(gzip_file, 'r') as gzip_in:
            json_var = json.loads(gzip_in.read().decode('utf-8'))
    except:
        raise
    return json_var

# save variable as JSON into .json.gz file
def gzip_save_var(gzip_file:str, save_var:Any) -> bool:
    """
    Save variable into .json.gz file (arbitrary extension)

    Parameters
    ----------
    gzip_file : str
        Target filename for .json.gz content
    var : Any
        JSON dumpable variable
    
    Returns
    -------
    success : bool
        True (otherwise raises exception!)
    """
    try:
        json_bytes = (json.dumps(save_var) + "\n").encode('utf-8')
        with gzip.GzipFile(gzip_file, 'w') as gzip_out:
            gzip_out.write(json_bytes)
        return True
    except:
        raise

# authentication
def isic_auth_token(base_url:str, username:str, password:str) -> str:
    """
    Makes a login requests and returns the Girder-Token header.
    
    Parameters
    ----------
    base_url : str
        Fully qualified hostname + API URI, e.g. https://host/api/v0
    username : str
        Username to log into the API (elem[0] in auth=() tuple in .get(...))
    password : str
        Password to pass into the API (elem[1] in auth=() tuple in .get(...))
    
    Returns
    -------
    auth_token : str
        Authentication (Girder-) Token (if successful, otherwise None)
    """
    auth_response = requests.get(make_url(base_url, 'user/authentication'),
                                 auth=(username, password))
    if not auth_response.ok:
        warnings.warn('Login error: ' + auth_response.json()['message'])
        return None
    return auth_response.json()['authToken']['token']

# URL generation
def make_url(base_url:str, endpoint:str) -> str:
    """
    Concatenates the base_url with '/' and the endpoint.
    
    Parameters:
    -----------
    base_url : str
        Fully qualified hostname + API URI, e.g. https://host/api/v0
    endpoint : str
        Endpoint in the API, e.g. study, dataset, or image
    """
    return base_url + '/' + endpoint

def object_pretty(obj:object, p:object, cycle:bool = False, fields:list = None):
    if fields is None:
        return
    t = str(type(obj)).replace('<class \'', '').replace('\'>', '')
    if cycle:
        p.text(t + '(id=' + getattr(obj, 'id') + ')')
        return
    with p.group(4, t + '({', '})'):
        if isinstance(fields, list):
            for field in fields:
                p.breakable()
                if not '.' in field:
                    field_val = getattr(obj, field)
                else:
                    field_val = getxattr(obj, field)
                if isinstance(field_val, str):
                    p.text('\'' + field + '\': \'' + field_val + '\',')
                elif isinstance(field_val, dict):
                    p.text('\'' + field + '\': { ... dict with ' +
                        str(len(field_val)) + ' fields},')
                elif isinstance(field_val, list):
                    p.text('\'' + field + '\': [ ... list with ' +
                        str(len(field_val)) + ' items],')
                else:
                    field_str = str(field_val)
                    if len(field_str) > 60:
                        field_str = field_str[:27] + ' ... ' + field_str[-27:]
                    p.text('\'' + field + '\': ' + field_str + ',')
        elif isinstance(fields, dict):
            for name, field in fields.items():
                p.breakable()
                if not '.' in field:
                    field_val = getattr(obj, field)
                else:
                    field_val = getxattr(obj, field)
                if isinstance(field_val, str):
                    p.text('\'' + name + '\': \'' + field_val + '\',')
                elif isinstance(field_val, dict):
                    p.text('\'' + name + '\': { ... dict with ' +
                        str(len(field_val)) + ' fields},')
                elif isinstance(field_val, list):
                    p.text('\'' + name + '\': [ ... list with ' +
                        str(len(field_val)) + ' items],')
                else:
                    field_str = str(field_val)
                    if len(field_str) > 60:
                        field_str = field_str[:27] + ' ... ' + field_str[-27:]
                    p.text('\'' + name + '\': ' + field_str + ',')
        else:
            raise ValueError('Invalid list of fields.')

# decode image superpixel
def superpixel_index(rgb_array:numpy.ndarray) -> numpy.ndarray:
    """
    Decode an RGB representation of a superpixel label into its native scalar
    value.

    Parameters
    ----------
    rgb_array : 3d numpy.ndarray (or imageio.core.util.Array)
        Image content of a ISIC superpixel PNG (RGB-encoded values)
    as_uint16 : bool
        By default, this function will only consider the first 2 planes!
    
    Returns
    -------
    superpixel_index : 2d numpy.ndarray
        2D Image (uint16) with superpixel indices
    """
    return (rgb_array[..., 0].astype(numpy.uint32) + 
        (rgb_array[..., 1].astype(numpy.uint32) << numpy.uint32(8)) +
        (rgb_array[..., 2].astype(numpy.uint32) << numpy.uint32(16))).astype(numpy.int32)

# create superpixel -> pixel index array
@numba.jit('i4[:,:](i4[:,:])', nopython=True) # (numba.int32[:,:](numba.int32[:,:]), nopython=True)
def superpixel_decode_img(pixel_img):
    """
    Decode a superpixel (patch) image to a dictionary with (1D) coordinates.

    Parameters
    ----------
    idx_array : 2d numpy.ndarray
        Image with superpixel index in each pixel
    
    Returns
    -------
    superpixel_map : dict
        Dict which maps from superpixel index (0-based) to 1D coordinates
        in the original (flattened) image space.
    """
    pixel_flat = pixel_img.flatten()
    spcounts = numpy.bincount(pixel_flat)
    spcountmax = numpy.amax(spcounts).item() + 1
    sp_to_p = numpy.zeros(len(spcounts) * spcountmax,
        dtype=numpy.int32).reshape(len(spcounts), spcountmax)
    spcounts = numpy.zeros(len(spcounts), dtype=numpy.int32)
    for idx in range(pixel_flat.size):
        pixel_val = pixel_flat[idx]
        sp_to_p[pixel_val, spcounts[pixel_val]] = idx
        spcounts[pixel_val] += 1
    for idx in range(len(spcounts)):
        sp_to_p[idx, -1] = spcounts[idx]
    return sp_to_p

# URI encode
_uri_letters = ' !"#$%&\'()*+,/:;<=>?@[\\]^`{|}~'
def uri_encode(uri:str) -> str:
    letters = ['%' + hex(ord(c))[-2:] if c in _uri_letters else c for c in uri]
    return ''.join(letters)

@numba.jit(nopython=True)
def xplus1(value):
    newvalue = numpy.zeros(value.size, dtype=numpy.int32)
    for idx in range(value.size):
        newvalue[idx] = value[idx] + 1
    return newvalue



"""
isicarchive.IsicApi

This module provides the isicapi class/object to access the ISIC
archive programmatically.

To instantiate a connection object without credentials (public use),
simply create the object without arguments:

   >>> from isicarchive import IsicApi
   >>> api = IsicApi()

If your code has only access to the username, the password will
requested using the getpass.getpass(...) call internally:

   >>> api = IsicApi(username)

If on the other hand your code already has stored the password
internally, you can also pass it to the IsicApi call:

   >>> api = IsicApi(username, password)

By default, the class uses ``https://isic-archive.com`` as hostname
(including the https protocol!), and ``/api/v1`` as the API URI.
These parameters can be overriden, which is useful should the URL
change:

   >>> api = IsicApi(hostname='https://new-archive.com')

or

   >>> api = IsicApi(api_uri='/api/v2')

For additional information on the individual endpoints as exposed by
the archive (website) in its current form, please visit

http://isic-archive.com/api/v1

You'll find documentation and examples for each of the endpoints,
together with the ability to test them separately.

Methods
-------
annotation
    Create an annotation object or retrieve a list of annotations
annotation_list
    Yields a generator for annotation JSON dicts
dataset
    Create a dataset object or retrieve a list of datasets
dataset_list
    Yields a generator for dataset JSON dicts
image
    Create an image object or retrieve a list of images
image_list
    Yields a generator for image JSON dicts
study
    Create a study object or retrieve a list of studies
study_list
    Yields a generator for study JSON dicts
"""

__version__ = '0.4.0'


import copy
import os
import tempfile
import time
from typing import Any, Optional, Tuple
import warnings

import getpass
import imageio
import json
import requests

from . import func
from . import vars
from .annotation import Annotation
from .dataset import Dataset
from .image import Image
from .study import Study

_repr_pretty_list = {
    'base_url': '_base_url',
    'username': 'username',
    'cache_folder': '_cache_folder',
    'cached_images': '_image_cache',
    'loaded_datasets': 'datasets',
    'loaded_studies': 'studies',
    'obj_annotations': '_annotation_objs',
    'obj_datasets': '_dataset_objs',
    'obj_images': '_image_objs',
    'obj_studies': '_study_objs',
}

def _mangle_id_name(object_id:str, name:str) -> Tuple[str, str]:
    """
    (Un-) Mangle possible object ID and name

    Parameters
    ----------
    object_id : str
        Object ID (or None)
    name : str
        Name (dataset, image, or study)

    Returns (Tuple)
    -------
    object_id : str
        Object id (if it is one)
    name : str
        Name (from object id, if it isn't one!)
    """
    if object_id is None:
        object_id, name = name, object_id
    elif isinstance(object_id, dict):
        if '_id' in object_id:
            object_id = object_id['_id']
        elif 'id' in object_id:
            object_id = object_id['id']
    if (not object_id is None) and ((len(object_id) != 24)
        or (not func.could_be_mongo_object_id(object_id))):
        if name is None:
            object_id, name = name, object_id
        else:
            object_id = None
    return object_id, name

class IsicApi(object):
    """
    IsicApi

    Attributes
    ----------
    datasets : dict
        Resolving dataset names to unique mongodb ObjectIds (cache)
        This will be pre-loaded with all datasets, regardless of login
    images : dict
        Resolving image names to unique mongodb ObjectIds (cache)
    meta_hist : dict
        Returned JSON response from /image/histogram endpoint as dict
        This will be pre-loaded if a username is given
    studies : dict
        Resolving study names to unique mongodb ObjectIds (cache)
        This will be pre-loaded with all studies, regardless of login
    
    Methods
    -------
    annotation(object_id=None, params=None)
        Retrieve one annotation (object) or annotations (list)
    dataset(object_id=None, name=None, params=None)
        Retrieve one dataset (object) or datasets (list)
    image(object_id=None, name=None, params=None)
        Retrieve one image (object) or images (list)
    study(object_id=None, name=None, params=None)
        Retrieve one study (object) or studies (list)
    """


    def __init__(self,
        username:Optional[str] = None,
        password:Optional[str] = None,
        hostname:Optional[str] = None,
        api_uri:Optional[str] = None,
        cache_folder:Optional[str] = None,
        ):

        """IsicApi.__init__: please refer to IsicApi docstring!"""
        
        # Check input arguments
        if hostname is None or hostname == '':
            hostname = vars.ISIC_BASE_URL
        elif (len(hostname)) < 8 or (hostname[0:8].lower() != 'https://'):
            if hostname[0:7].lower() == 'http://':
                raise ValueError('IsicApi must use https:// as protocol!')
            hostname = 'https://' + hostname

        if api_uri is None:
            api_uri = vars.ISIC_API_URI
        elif api_uri == '' or api_uri[0] != '/':
            api_uri = '/' + api_uri

        # Prepare object:
        # _TYPEs: map id -> PARTIAL JSON dict (containing name!)
        # _TYPE_objs: map id -> created Object (after download)
        # TYPEs: map name -> id (for images only when downloaded)
        self._api_uri = api_uri
        self._annotation_objs = dict()
        self._auth_token = None
        self._base_url = hostname + api_uri
        self._cache_folder = None
        self._current_annotation = None
        self._current_dataset = None
        self._current_image = None
        self._current_study = None
        self._datasets = dict()
        self._dataset_objs = dict()
        self._hostname = hostname
        self._image_cache = dict()
        self._image_cache_last = '0' * 24
        self._image_cache_timeout = 0.0
        self._image_objs = dict()
        self._studies = dict()
        self._study_objs = dict()
        self._temp_file = None
        self.datasets = dict()
        self.images = dict()
        self.meta_hist = dict()
        self.studies = dict()
        self.username = None

        # accept cache folder?
        if cache_folder and os.path.exists(cache_folder):
            if os.path.isdir(cache_folder):
                try:
                    self._temp_file = tempfile.TemporaryFile(dir=cache_folder)
                    self._cache_folder = cache_folder
                    for sf in '0123456789abcdef':
                        cs_folder = cache_folder + os.sep + sf
                        if not os.path.exists(cs_folder):
                            os.mkdir(cs_folder)
                except:
                    warnings.warn('Error creating a file in ' + cache_folder)

        # Login required
        if username is not None:
            self.username = username

            # And get the password using getpass
            if password is None:
                password = getpass.getpass('Password for "%s":' % (username))

            # Login
            self._auth_token = func.isic_auth_token(
                self._base_url, username, password)

            # if login succeeded, collect meta information histogram
            if self._auth_token:
                self.meta_hist = func.get(self._base_url, 'image/histogram',
                    self._auth_token).json()

        # pre-populate information about datasets and studies
        items = self.dataset(params={'limit': 0, 'detail': 'true'})
        for item in items:
            self._datasets[item['_id']] = item
            self.datasets[item['name']] = item['_id']
        items = self.study(params={'limit': 0, 'detail': 'true'})
        for item in items:
            self._studies[item['_id']] = item
            self.studies[item['name']] = item['_id']

    # output
    def __repr__(self) -> str:
        return 'IsicApi(\'%s\', None, \'%s\', \'%s\', \'%s\')' % (
            self.username, self._hostname, self._api_uri, self._cache_folder)
    def __str__(self) -> str:
        if self._auth_token:
            return 'IsicApi logged into {0:s} with user {1:s}.'.format(
                self.username, self._hostname)
        return 'IsicApi accessing %s.' % self._hostname
    def _repr_pretty_(self, p:object, cycle:bool = False):
        func.object_pretty(self, p, cycle, _repr_pretty_list)

    # Annotation endpoint
    def annotation(self,
        object_id:str = None,
        params:dict = None,
        ) -> any:
        """
        annotation endpoint, allows to
        - retrieve information about multiple annotations
        - retrieve one specific annotation (object)

        Parameters
        ----------
        object_id : str
            Valid 24-character mongodb objectId for the annotation
        params : dict
            Optional parameters for the GET request
        
        Returns
        -------
        annotation : object
            For a single annotation, returns an Annotation object
        annotations : list
            For multiple annotations, a list of JSON dicts
        """
        object_id = _mangle_id_name(object_id, 'null')
        if isinstance(object_id, tuple):
            object_id = object_id[0]
        if object_id is None:
            if not isinstance(params, dict):
                raise ValueError(
                    'Annotation list requires a params dict with a study_id.')
            if 'study_id' in params:
                params = copy.copy(params)
                params['studyId'] = params['study_id']
                del params['study_id']
            if not 'studyId' in params:
                raise ValueError(
                    'Annotation list requires field study_id in params.')
            return func.get(self._base_url,
                'annotation', self._auth_token, params).json()
        if not func.could_be_mongo_object_id(object_id):
            raise ValueError('Invalid objectId format of object_id parameter.')
        if object_id in self._annotation_objs:
            self._current_annotation = self._annotation_objs[object_id]
            return self._current_annotation
        annotation = func.get(self._base_url,
            'annotation/' + object_id, self._auth_token, params).json()
        if not '_id' in annotation:
            raise KeyError('Annotation with id %s not found.' % (object_id))
        annotation_obj = Annotation(annotation, api=self)
        self._annotation_objs[annotation['_id']] = annotation_obj
        self._current_annotation = annotation_obj
        return annotation_obj

    def annotation_list(self, params:dict = None, as_object:bool = False) -> iter:
        """
        Annotation list/iterator

        Parameters
        ----------
        params : dict
            Dictionary with GET parameters for the query string, must
            contain 'study' (name) or 'study_id'
        as_object : bool
            If set to false (default), yields dicts, otherwise objects
        
        Yields
        ------
        object
            Annotation as JSON dict or object
        """
        if isinstance(params, str):
            params = {'study': params}
        elif isinstance(params, dict):
            params = copy.copy(params)
        else:
            raise ValueError('Invalid or missing params.')
        if 'studyId' in params:
            params['study'] = params['studyId']
        elif 'study_id' in params:
            params['study'] = params['study_id']
            del params['study_id']
        if not func.could_be_mongo_object_id(params['study']):
            if params['study'] in self.studies:
                params['study'] = self.studies[params['study']]
            else:
                raise ValueError('Unknown study: ' + params['study'])
        params['studyId'] = params['study']
        del params['study']
        if as_object:
            params['detail'] = 'true'
        items = self.annotation(params=params)
        for item in items:
            if as_object:
                yield Annotation(from_json=item, api=self)
            else:
                yield item

    # cache image information
    def _cache_images(self, from_list:dict):
        for item in from_list:
            if not item['_id'] in self._image_cache:
                self._image_cache[item['_id']] = item
    def cache_images(self):
        """
        Create or update the local image details cache file

        No input parameters and no return value.
        """
        if not self._cache_folder or (not os.path.isdir(self._cache_folder)):
            return
        if self._image_cache_timeout >= time.time():
            return
        image_cache_filename = func.cache_filename('0' * 24,
            'imcache', '.json.gz', api=self)
        if os.path.exists(image_cache_filename):
            try:
                self._image_cache = func.gzip_load_var(image_cache_filename)
                self._image_cache_last = sorted(self._image_cache.keys())[-1]
            except:
                os.remove(image_cache_filename)
                warnings.warn('Invalid image cache file.')
        initial_offset = 0
        limit = vars.ISIC_IMAGES_PER_CACHING
        limit_half = limit // 2
        offset = 0
        params = {
            'detail': 'true',
            'limit': str(limit),
            'offset': str(offset),
            'sort': '_id',
            'sortdir': '1',
        }
        num_loaded = len(self._image_cache)
        last_id = self._image_cache_last
        if num_loaded > 0:
            initial_offset = max(0, num_loaded - limit_half)
            offset = initial_offset
        params['offset'] = str(offset)
        partial_list = self.image(params=params)
        self._cache_images(partial_list)
        partial_list_ids = sorted(map(lambda item: item['_id'], partial_list))
        while partial_list_ids[0] > last_id and (offset > 0):
            offset = max(0, offset - limit)
            params['offset'] = str(offset)
            partial_list = self.image(params=params)
            self._cache_images(partial_list)
            partial_list_ids = sorted(map(lambda item: item['_id'], partial_list))
        offset = initial_offset
        while len(partial_list) == limit:
            offset += limit
            params['offset'] = str(offset)
            partial_list = self.image(params=params)
            self._cache_images(partial_list)
        try:
            if num_loaded < len(self._image_cache):
                func.gzip_save_var(image_cache_filename, self._image_cache)
                self._image_cache_last = sorted(self._image_cache.keys())[-1]
        except:
            warnings.warn('Error writing cache file.')
        self._image_cache_timeout = time.time() + vars.ISIC_IMAGE_CACHE_UPDATE_LASTS

    # Generic /dataset endpoint
    def dataset(self,
        object_id:str = None,
        name:str = None,
        params:dict = None,
        ) -> any:
        """
        dataset endpoint, allows to
        - retrieve information about all available datasets
        - retrieve one specific dataset (object)

        Parameters
        ----------
        object_id : str
            Valid 24-character mongodb objectId for the dataset
        name : str
            Alternatively the name of the dataset
        params : dict
            Optional parameters for the GET request
        
        Returns
        -------
        dataset : object
            For a single dataset, returns a Dataset object
        datasets : list
            For multiple datasets, a list of JSON objects
        """
        (object_id, name) = _mangle_id_name(object_id, name)
        if object_id is None:
            if name is None:
                return func.get(self._base_url,
                    'dataset', self._auth_token, params).json()
            try:
                if name in self.datasets:
                    object_id = self.datasets[name]
                else:
                    raise KeyError('Dataset "%s" not found.' % (name))
            except:
                raise
        if not func.could_be_mongo_object_id(object_id):
            raise ValueError('Invalid object_id format.')
        if object_id in self._dataset_objs:
            self._current_dataset = self._dataset_objs[object_id]
            return self._current_dataset
        if object_id in self._datasets:
            dataset = self._datasets[object_id]
        else:
            try:
                dataset = func.get(self._base_url,
                    'dataset/' + object_id, self._auth_token, params).json()
            except:
                raise
        if not '_id' in dataset:
            raise KeyError('Dataset with id %s not found.' % (object_id))
        if not dataset['name'] in self.datasets:
            self.datasets[dataset['name']] = dataset['_id']
        dataset_obj = Dataset(dataset, api=self)
        self._dataset_objs[dataset['_id']] = dataset_obj
        self._current_dataset = dataset_obj
        return dataset_obj

    # Dataset list (generator)
    def dataset_list(self, params:dict = None, as_object:bool = False) -> iter:
        """
        Dataset list/iterator

        Parameters
        ----------
        params : dict
            Optional GET parameters for the query string
        as_object : bool
            If set to false (default), yields dicts, otherwise objects
        
        Yields
        ------
        object
            Dataset as JSON dict or object
        """
        if as_object:
            if isinstance(params, dict):
                params = copy.copy(params)
            else:
                params = dict()
            params['detail'] = 'true'
        datasets = self.dataset(params=params)
        for dataset in datasets:
            if as_object:
                yield Dataset(from_json=dataset, api=self)
            else:
                yield dataset

    # find images
    def find_images(self, filter_spec:dict) -> Any:
        if not self._cache_folder:
            raise ValueError('Requires cache folder to be set.')
        self.cache_images()
        pass

    # Generic /image endpoint
    def image(self,
        object_id:str = None,
        name:str = None,
        params:dict = None,
        save_as:str = None,
        load_imagedata:bool = False,
        load_superpixels:bool = False,
        ) -> any:
        """
        image endpoint, allows to
        - retrieve information about images
        - retrieve (or download) one image (object)

        Parameters
        ----------
        object_id : str
            valid 24-character mongodb objectId for the dataset
        name : str
            alternatively the name of the dataset
        params : dict
            optional parameters for the GET request
        save_as : str
            Optional filename for the image to be saved as, in which
            case nothing is returned and the images is ONLY downloaded!
        
        Returns
        -------
        object
            for a single dataset, returns a JSON object with _id and name
        list
            for multiple datasets, a list of JSON objects
        """
        (object_id, name) = _mangle_id_name(object_id, name)
        if object_id is None:
            if name is None:
                return func.get(self._base_url,
                    'image', self._auth_token, params).json()
            try:
                if name in self.images:
                    object_id = self.images[name]
                else:
                    object_id = func.get(self._base_url,
                        'image', self._auth_token, params={'name': name}).json()
                    if isinstance(object_id, list):
                        object_id = object_id[0]
                    if '_id' in object_id:
                        object_id = object_id['_id']
                        self.images[name] = object_id
                    else:
                        raise KeyError('Image %s not found' % (name))
            except:
                raise
        if not func.could_be_mongo_object_id(object_id):
            raise ValueError('Invalid object_id format.')
        if not save_as is None:
            try:
                func.get(self._base_url,
                    'image/' + object_id + '/download',
                    self._auth_token, params, save_as)
            except:
                raise
            return
        if save_as is None and object_id in self._image_objs:
            self._current_image = self._image_objs[object_id]
            if load_imagedata and self._current_image.data is None:
                self._current_image.load_imagedata()
            if load_superpixels and self._current_image.superpixels['idx'] is None:
                self._current_image.load_superpixels()
            return self._current_image
        if object_id in self._image_cache:
            image = self._image_cache[object_id]
        else:
            try:
                image = func.get(self._base_url,
                    'image/' + object_id, self._auth_token, params).json()
            except:
                raise
        if not '_id' in image:
            raise KeyError('Image with id %s not found.' % (object_id))
        if not image['name'] in self.images:
            self.images[image['name']] = image['_id']
        image_obj = Image(image, api=self,
            load_imagedata=load_imagedata, load_superpixels=load_superpixels)
        self._image_objs[image['_id']] = image_obj
        self._current_image = image_obj
        return image_obj

    # Image list (generator)
    def image_list(self, params:dict = None, as_object:bool = False) -> iter:
        """
        Image list/iterator

        Parameters
        ----------
        params : dict
            Optional GET parameters for the query string
        as_object : bool
            If set to false (default), yields dicts, otherwise objects
        
        Yields
        ------
        object
            Image as JSON dict or object
        """
        if as_object:
            if isinstance(params, dict):
                params = copy.copy(params)
            else:
                params = dict()
            params['detail'] = 'true'
        images = self.image(params=params)
        for image in images:
            if as_object:
                yield Image(image, api=self)
            else:
                yield image

    # list datasets (printed)
    def list_datasets(self):
        print('List of datasets available at ' + self._hostname + ':')
        for name, object_id in self.datasets.items():
            if object_id in self._dataset_objs:
                dataset_obj = self._dataset_objs[object_id]
            else:
                dataset_obj = self.dataset(self._datasets[object_id])
            print(' * ' + name + ' (id=' + object_id + ')')
            print('     - ' + dataset_obj.description + ' (description)')
            print('     - {0:d} approved images'.format(dataset_obj.count))
            print('     - {0:d} images from cache'.format(
                len(dataset_obj.images)))
            print('     - {0:d} images for review'.format(
                len(dataset_obj.images_for_review)))

    # list studies (printed)
    def list_studies(self):
        print('List of studies available at ' + self._hostname + ':')
        for name, object_id in self.studies.items():
            if object_id in self._study_objs:
                study_obj = self._study_objs[object_id]
            else:
                study_obj = self.study(self._studies[object_id])
            print(' * ' + name + ' (id=' + object_id + ')')
            print('     - ' + study_obj.description + ' (description)')
            print('     - {0:d} annotations ({1:d} loaded)'.format(
                len(study_obj.annotations), len(study_obj._obj_annotations)))
            print('     - {0:d} features'.format(
                len(study_obj.features)))
            print('     - {0:d} images ({1:d} loaded)'.format(
                len(study_obj.images), len(study_obj._obj_images)))
            print('     - {0:d} questions'.format(
                len(study_obj.questions)))

    # Generic /study endpoint
    def study(self,
        object_id:str = None,
        name:str = None,
        params:dict = None,
        ) -> any:
        """
        study endpoint, allows to
        - retrieve information about all available studies
        - retrieve one specific study (object)

        Parameters
        ----------
        object_id : str
            Valid 24-character mongodb objectId for the study
        name : str
            Alternatively the name of the study
        params : dict
            Optional parameters for the GET request
        
        Returns
        -------
        study : object
            For a single study, returns a Study object
        studies : list
            For multiple studies, a list of JSON dicts
        """
        (object_id, name) = _mangle_id_name(object_id, name)
        if object_id is None:
            if name is None:
                return func.get(self._base_url,
                'study', self._auth_token, params=params).json()
            try:
                if name in self.studies:
                    object_id = self.studies[name]
                else:
                    raise KeyError('Study "%s" not found.' % (name))
            except:
                raise
        if not func.could_be_mongo_object_id(object_id):
            raise ValueError('Invalid object_id format.')
        if object_id in self._study_objs:
            return self._study_objs[object_id]
        if object_id in self._studies:
            study = self._studies[object_id]
        else:
            try:
                study = func.get(self._base_url,
                    'study/' + object_id, self._auth_token, params).json()
            except:
                raise
        if not '_id' in study:
            raise KeyError('Dataset with id %s not found.' % (object_id))
        if not study['name'] in self.studies:
            self.studies[study['name']] = study['_id']
        study_obj = Study(study, api=self)
        self._study_objs[study['_id']] = study_obj
        return study_obj
    
    # study list (generator)
    def study_list(self, params:dict = None, as_object:bool = False) -> iter:
        """
        Study list/iterator

        Parameters
        ----------
        params : dict
            Optional GET parameters for the query string
        as_object : bool
            If set to false (default), yields dicts, otherwise objects
        
        Yields
        ------
        object
            Study as JSON dict or object
        """
        if as_object:
            if isinstance(params, dict):
                params = copy.copy(params)
            else:
                params = dict()
            params['detail'] = 'true'
        else:
            if not isinstance(params, dict):
                params = {'detail': 'false'}
            elif not 'detail' in params:
                params = copy.copy(params)
                params['detail'] = 'false'
        studies = self.study(params=params)
        for study in studies:
            if as_object:
                yield Study(from_json=study, api=self)
            else:
                yield study
